from contextlib import contextmanager
import sys, os


@contextmanager
def cd(path=None):
    from .config import Config
    path = path or Config.iot49_dir()
    cwd = os.getcwd()
    os.chdir(os.path.expandvars(os.path.expanduser(path)))
    try:
        yield
    finally:
        os.chdir(cwd)


@contextmanager
def redirect_stdout_stderr(out, err):
    # temporarily redirect stdout, stderr
    old_stdout = sys.stdout
    old_stderr = sys.stderr
    sys.stdout = out
    sys.stderr = err
    try:
        yield
    finally:
        sys.stdout = old_stdout
        sys.stderr = old_stderr



from .eval import RemoteError
from .eval_file_ops import EvalFileOps
from datetime import datetime
from termcolor import colored

import time, os, logging

logger = logging.getLogger(os.path.splitext(os.path.basename(__file__))[0])

class EvalRlist(EvalFileOps):
    """Add file listing capabilities"""

    def rlist(self, path='/', data_consumer=None, show=False):
        """Return and optionally display files stored on remote"""
        out = RlistOutput(data_consumer, show)
        try:
            self._remote_exec(f"rlist({repr(path)})", _rlist_func, out.data_consumer)
        except RemoteError as e:
            raise
        return out.files


###############################################################################
# Collect output from _mcu_list

class TZ:
    # convert micropython localtime (e.g. mtime) to gmtime

    __ts = time.time()
    __localtime_gmtime = (datetime.fromtimestamp(__ts) - datetime.utcfromtimestamp(__ts)).total_seconds()

    @staticmethod
    def local2gmtime(local_time):
        return local_time - TZ.__localtime_gmtime

    @staticmethod
    def gmtime2local(gm_time):
        return gm_time + TZ.__localtime_gmtime


class RlistOutput(TZ):

    def __init__(self, data_consumer=None, show=False):
        # show: prints file list to output
        # (otherwise just progress, if output not None)
        self._output = data_consumer
        self._show = show
        self._level_offset = 0
        self._path_stack = []
        self._files = {}
        self._line_buffer = bytes()

    @property
    def files(self):
        return self._files

    def _indent(self, level):
        return ' '*4*(level + self._level_offset)

    def data_consumer(self, b):
        # b could be any fragment or combination of lines!
        # Drop \x04 returned after softreset (???)
        self._line_buffer += b.replace(b'\x04', b'')
        while b'\r\n' in self._line_buffer:
            line, self._line_buffer = self._line_buffer.split(b'\r\n', 1)
            kind, level, path, mtime, size = line.split(b',')
            path = eval(path)
            if len(path)<=0: continue
            level = int(level)
            size  = int(size)
            mtime = int(self.local2gmtime(int(mtime)))
            mtime_fmt = datetime.fromtimestamp(mtime).strftime("%b %d %H:%M %Y")
            full_path = os.path.join(*self._path_stack[:level], path)
            if kind == b'D':
                self._files[full_path] = (mtime, -1)
                while len(self._path_stack) < level+1:
                    self._path_stack.append('')
                self._path_stack[level] = path
                if level != 0:
                    if self._show and self._output:
                        path += '/'
                        self._output(f"{' ':7}  {' ':18} {self._indent(level)}{colored(path, 'green')}\n")
                else:
                    self._level_offset = -1
            else:
                self._files[full_path] = (mtime, size)
                if self._output:
                    if self._show:
                        self._output(f"{int(size):7}  {mtime_fmt:18} {self._indent(level)}{colored(path, 'blue')}\n")
                    elif len(self._files) > 50 and len(self._files) % 10 == 0:
                        # show progress
                        self._output('.')


###############################################################################
# code snippet (runs on remote)

_rlist_func = """
t_off = 0
try:
    import machine
    t_off = 946684800
except ImportError:
    pass

import os
def rlist(path, level=0):
    stat = os.stat(path)
    fsize = stat[6]
    mtime = stat[7] + t_off
    if stat[0] & 0x4000:
        up = os.getcwd()
        os.chdir(path)
        print("D,{},{},{},-1".format(level, repr(path), mtime))
        for p in sorted(os.listdir()):
            if p.startswith('.'): continue
            rlist(p, level+1)
        os.chdir(up)
    else:
        print("F,{},{},{},{}".format(level, repr(path), mtime, fsize))
"""



from .discover import Discover

from zeroconf import Zeroconf, ServiceBrowser, ServiceStateChange
import os, socket, logging, time

logger = logging.getLogger(os.path.splitext(os.path.basename(__file__))[0])


class DiscoverMdns(Discover):
    """zeroconf device discovery"""

    def __init__(self, scan_rate:float=5):
        """Start a daemon thread that continually scans ports every scan_rate seconds."""
        super().__init__()

    def scan(self):
        # return url's of devices that are online
        self.urls = set()
        zc = Zeroconf()
        try:
            ServiceBrowser(zc, "_mp._tcp.local.", self)
            ServiceBrowser(zc, "_ws._tcp.local.", self)
            ServiceBrowser(zc, "_telnet._tcp.local.", self)
            # give the scanner some time to find everything ...
            time.sleep(1)
        except Exception as ex:
            print(ex)
            logger.exception(f"Unhandled exception in DiscoverMdns._scanner")
        finally:
            zc.close()
        return self.urls

    def _url(self, zeroconf, service_type, name):
        info = zeroconf.get_service_info(service_type, name)
        if info == None: return
        ip = socket.inet_ntoa(info.addresses[0])
        port = info.port
        scheme = service_type.split('.')[0][1:]
        url = f"{scheme}://{ip}:{port}"
        return url

    def add_service(self, zeroconf, service_type, name):
        self.urls.add(self._url(zeroconf, service_type, name))

    def remove_service(self, zeroconf, service_type, name):
        pass

    def update_service(self, zeroconf, service_type, name):
        pass



from .pyboard import Pyboard, PyboardError

import logging, time, os

logger = logging.getLogger(os.path.splitext(os.path.basename(__file__))[0])

RAW_REPL_MSG = b"raw REPL; CTRL-B to exit\r\n"

MCU_RAW_REPL      = b'\x01'    # ctrl-A enter raw repl
MCU_FRIENDLY_REPL = b'\x02'    # ctrl-B enter friendly repl
MCU_ABORT         = b'\x03'    # ctrl-C abort
MCU_RESET         = b'\x04'    # ctrl-D softreset
MCU_EVAL          = b'\r\x04'  # start evaluation (raw repl)
EOT               = b'\x04'
CR                = b'\r'


class Pydevice(Pyboard):

    def __init__(self, device):
        """Pyboard with a different initializer."""
        if not hasattr(device, 'use_raw_paste'):
            device.use_raw_paste = True
        self.use_raw_paste = device.use_raw_paste
        self.serial = device

    def read_until(self, min_num_bytes, ending, timeout=10, data_consumer=None):
        # patch: don't call data_consumer on single chars (if more data is available)
        #        kernal print is very slow on Pi
        # if data_consumer is used then data is not accumulated and the ending must be 1 byte long
        assert data_consumer is None or len(ending) == 1
        data = self.serial.read(min_num_bytes)
        timeout_count = 0
        while True:
            if data.endswith(ending):
                if data_consumer:
                    data_consumer(data)
                break
            if self.serial.inWaiting() > 0:
                new_data = self.serial.read(1)
                data = data + new_data
                timeout_count = 0
            else:
                if data_consumer and len(data) > 0:
                    data_consumer(data)
                    data = b''
                timeout_count += 1
                if timeout is not None and timeout_count >= 100 * timeout:
                    break
                time.sleep(0.01)
        return data

    def enter_raw_repl(self):
        self.serial.write(b"\r\x03\x03")  # ctrl-C twice: interrupt any running program

        # flush input (without relying on serial.flushInput())
        n = self.serial.inWaiting()
        while n > 0:
            self.serial.read(n)
            n = self.serial.inWaiting()

        self.serial.write(b"\r\x01")  # ctrl-A: enter raw REPL
        data = self.read_until(1, RAW_REPL_MSG)
        if not data.endswith(RAW_REPL_MSG):
            print(data)
            raise PyboardError(f"could not enter raw repl:\n  expected '{RAW_REPL_MSG}'\n  got '{data}'")

    def exec(self, command, data_consumer=None):
        self.exec_raw_no_follow(command)
        ret, ret_err = self.follow(timeout=None, data_consumer=data_consumer)
        if ret_err:
            raise PyboardError("exception", ret, ret_err)
        return ret

    def softreset(self):
        device = self.serial
        device.write(MCU_ABORT)
        device.write(MCU_RESET)
        device.write(CR)
        data = self.read_until(1, RAW_REPL_MSG)
        if not data.endswith(RAW_REPL_MSG):
            raise PyboardError(f"could not reset board:\n  expected '{RAW_REPL_MSG}'\n  got '{data}'")



from .eval import RemoteError
from .eval_rsync import EvalRsync
import socket, os, struct, time


EOT = b'\x04'

class MpProtocol(EvalRsync):

    def __init__(self, device):
        assert device.scheme == "mp"
        super().__init__(device)

    # implement abstract exec
    def exec(self, code: str, data_consumer=None) -> bytes:
        if len(code) == 0: return
        if isinstance(code, str): code = code.encode()
        self.sendall(b'evex\n')
        self.sendall(str(len(code)).encode())
        self.sendall(b'\n')
        self.sendall(code)
        # compilation may take a long time!
        ok = self.readline(60)
        if ok != 'OK': raise RemoteError(f"eval: expected OK, got {ok}")
        # read result
        res = self.read_until_eot(data_consumer)
        err = self.read_until_eot(None)
        if len(err) > 0:
            raise RemoteError("", res, err)
        return res

    def fget(self, src, dst, chunk_size=-1):
        self.sendall(b'fget\n')
        self.sendall(src.encode())
        self.sendall(b'\n')
        ok = self.readline()
        if ok != 'OK': raise RemoteError(f"fget: expected OK, got {ok}")
        sz = int(self.readline())
        n = 0
        with open(dst, 'wb') as f:
            n = 0
            while n < sz:
                b = self.recv(min(1024, sz-n))
                f.write(b)
                n += len(b)

    # override ExecFileOps
    def fput(self, src, dst):
        sz = os.path.getsize(src)
        self.sendall(b"fput\n")
        self.sendall(dst.encode())
        self.sendall(b'\n')
        self.sendall(str(sz).encode())
        self.sendall(b'\n')
        with open(src, 'rb') as f:
            self.sendall(f.read())
        ok = self.readline(timeout=10)
        if ok != 'OK': raise RemoteError(f"fput: expected OK, got {ok}")

    def softreset(self):
        raise RemoteError("softreset not implemented for 'mp' protocol")
        # programmatic reset
        # self.exec(_softreset)

    def abort(self):
        # abort program execution - now sure how to do this?
        # perhaps disconnecting? interrupt?
        pass

    def recv(self, sz=1):
        return self.device.sock.recv(sz)

    def sendall(self, data):
        return self.device.sock.sendall(data)

    def readline(self, timeout=1):
        return self.device.readline(timeout)

    def read_until_eot(self, data_consumer):
        res = b""
        while True:
            try:
                b = self.recv(1)
            except socket.timeout:
                # print("MpProtocol.read_until_eof timed out! (ignored)")
                b = b""
            if b == EOT: return res
            if data_consumer: data_consumer(b)
            res += b



_softreset = """\
try:
    import microcontroller
    microcontroller.reset()
except ImportError:
    import machine
    machine.soft_reset()
"""



from .eval import RemoteError
from .eval_defaults import EvalDefaults
import logging, time, os, io, ast


logger = logging.getLogger(os.path.splitext(os.path.basename(__file__))[0])


class EvalFileOps(EvalDefaults):
    """Add file system operations"""

    def makedirs(self, path:str):
        """Make all directories required for path. No-op if directories exist."""
        self._remote_exec(f"makedirs({repr(path)})", _makedirs_func)

    def rm_rf(self, path:str, r:bool=True, f:bool=True):
        """rm -rf path"""
        self._remote_exec(f"rm_rf({repr(path)}, {repr(r)}, {repr(f)})", _rm_rf_func)

    def cat(self, path:str, data_consumer=None):
        """Show contents of path on console"""
        return self._remote_exec(f"cat({repr(path)})", _cat_func, data_consumer=data_consumer)

    def get_time(self):
        """Get struct time from mcu"""
        st = self._remote_exec(f"get_time()", _time_funcs)
        if not st:
            raise RemoteError(f"Cannot read time from {self.device.name}")
        st = eval(st.decode())
        if len(st) < 9:
            st += (-1, )
        return st

    def sync_time(self, tolerance:float=5):
        """Synchronize mcu time to host if they differ by more than tolerance seconds"""
        self._remote_exec(f"set_time({tuple(time.localtime())}, {tolerance})", _time_funcs)

    def fget(self, mcu_file:str, host_file:str, chunk_size:int=256):
        """Copy from microcontroller to host"""
        self.exec(f"f=open('{mcu_file}', 'rb')\nr=f.read")
        with open(host_file, 'wb') as f:
            while True:
                data = bytearray()
                data.extend(self.exec(f"print(r({chunk_size}))"))
                assert data.endswith(b"\r\n")
                try:
                    data = ast.literal_eval(str(data[:-2], "ascii"))
                    if not isinstance(data, bytes):
                        raise ValueError("Not bytes")
                except (UnicodeDecodeError, ValueError) as e:
                    raise RemoteError(f"fget: Could not interpret received data: {str(e)}")
                if not data: break
                f.write(data)
        self.exec("f.close()")

    def fput(self, host_file:str, mcu_file:str, chunk_size:int=256):
        """Copy from host to microcontroller"""
        self.makedirs(os.path.dirname(mcu_file))
        if not os.path.isfile(host_file): return
        self.exec(f"f=open('{mcu_file}','wb')\nw=f.write")
        with open(host_file, 'rb') as f:
            while True:
                data = f.read(chunk_size)
                if not data: break
                self.exec(f"w({repr(data)})")
        self.exec("f.close()")

    def _remote_exec(self, code:str, func:str, data_consumer=None) -> bytes:
        """Execute code on remote; upload code if required"""
        try:
            logger.debug(f"_remote_exec({code})")
            return self.exec(f"exec({repr(code)}, __iot49__)", data_consumer)
        except (RemoteError, OSError):
            # upload __iot49__ and try again
            logger.debug(f"_remote_exec: upload {func}")
            self.exec("if not '__iot49__' in globals(): __iot49__ = {}")
            self.exec(f"exec({repr(func)}, __iot49__)")
            logger.debug("_remote_exec: 2nd try")
            return self.exec(f"exec({repr(code)}, __iot49__)", data_consumer)


###############################################################################
# code snippets (run on remote)

_makedirs_func = """
import os
def makedirs(path):
    try:
        os.mkdir(path)
    except OSError as e:
        if e.args[0]==2:
            makedirs(path[:path.rfind('/')])
            os.mkdir(path)
        elif e.args[0]==17:
            pass
        else:
            raise
"""

_rm_rf_func = """
import os
def rm_rf(path, r, f):
    mode = os.stat(path)[0]
    if mode & 0x4000 != 0:
        if r:
            for file in os.listdir(path):
                rm_rf(path + '/' + file, r, f)
        if f:
            os.rmdir(path)
    else:
        os.remove(path)
"""

_cat_func = """
def cat(path):
    with open(path) as f:
        while True:
            line = f.readline()
            if not line:
                break
            print(line, end="")
"""

_time_funcs = """
import time
def get_time():
    print(tuple(time.localtime()), end="")

def set_time(st, tolerance=5):
    host  = time.mktime(st)
    local = time.time()
    # delete this comment, stops working with ws
    if abs(host-local) < tolerance:
        return
    try:
        import rtc
        rtc.RTC().datetime = st
    except ImportError:
        import machine as m
        st = list(st)
        st.insert(3, st[6])
        st[7] = 0
        m.RTC().datetime(st[:8])
"""



#!/usr/bin/env python3

from OpenSSL import crypto
from socket import gethostname

def create_key_cert_pair():
    # private key
    key = crypto.PKey()
    key.generate_key(crypto.TYPE_RSA, 1024)

    # signed certificate
    cert = crypto.X509()
    cert.get_subject().C  = "US"
    cert.get_subject().ST = "CA"
    cert.get_subject().L  = "Berkeley"
    cert.get_subject().O  = "IoT 49"
    cert.get_subject().OU = "IoT Python"
    cert.get_subject().CN = gethostname()
    cert.set_serial_number(1000)
    cert.gmtime_adj_notBefore(0)
    cert.gmtime_adj_notAfter(10*365*24*60*60)
    cert.set_issuer(cert.get_subject())
    cert.set_pubkey(key)
    cert.sign(key, 'sha1')

    return (
        crypto.dump_privatekey(crypto.FILETYPE_PEM, key),
        crypto.dump_certificate(crypto.FILETYPE_PEM, cert))


##########################################################################
# Example

def main():
    key, cert = create_key_cert_pair()
    print(key.decode())
    print(cert.decode())

if __name__ == "__main__":
    main()



from .device import Device
from .config import Config
from .eval import RemoteError
from .repl_protocol import ReplProtocol

from websocket import create_connection, WebSocketException
import os, time, logging, select

logger = logging.getLogger(os.path.splitext(os.path.basename(__file__))[0])

class WebreplDevice(Device):

    def __init__(self, url):
        self.use_raw_paste = False
        super().__init__(url)

    def inWaiting(self):
        r,_,_ = select.select([self.__ws], [], [], 0.1)
        return 1 if r else 0

    def read(self, size=1):
        res = b''
        while len(res) < size:
            r,_,_ = select.select([self.__ws], [], [], 0.1)
            if r: res += self.__ws.recv().encode()
            if len(res) >= size:
                return res
            time.sleep(0.01)

    def write(self, data):
        if len(data) < 252:
            return self.__ws.send(data)
        # slow down to avoid communication errors
        chunk_size = 64
        n = 0
        for i in range(0, len(data), chunk_size):
            logger.debug(f"webrepl_device.write({data[i:min(i+chunk_size, len(data))]})")
            n += self.__ws.send(data[i:min(i+chunk_size, len(data))])
            if n < len(data):
                time.sleep(0.5)
        return n

    def __enter__(self):
        try:
            self.__ws = create_connection(self.url, 3)
            self.__ws.settimeout(100)
            p = b'Password: '
            pp = self.read(len(p))
            self.write(Config.get_attr('webrepl_pwd', '?').encode())
            self.write(b'\r\n')
            return ReplProtocol(self)
        except WebSocketException as e:
            raise RemoteError(f"Websocket exception: {e}")

    def __exit__(self, type, value, traceback):
        try:
            # takes ages ...
            self.__ws.close()
        except Exception as e:
            logger.exception(f"***** Webrepdevice.close {e}")
        self.__ws = None



from .eval import Eval, RemoteError
import os, logging

logger = logging.getLogger(os.path.splitext(os.path.basename(__file__))[0])


class EvalDefaults(Eval):
    """Default implementations for select abstract methods of Eval"""

    @property
    def implementation(self):
        return self.exec("import sys; print(sys.implementation.name, end='')").decode()

    @property
    def platform(self):
        return self.exec("import sys; print(sys.platform, end='')").decode()



from .eval import RemoteError
from .eval_rlist import EvalRlist
from .utilities import cd
from .config import Config

from termcolor import colored
from glob import glob
from fnmatch import fnmatch
from collections import OrderedDict

import time, os, logging

logger = logging.getLogger(os.path.splitext(os.path.basename(__file__))[0])

"""File naming conventions:

Config.Package.files() returns:
    $IOT49 / path=package / name=file_path

Location of files on HOST & MCU:
    HOST:       $IOT49 / package / file_path
    MCU:        dest / file_path = full_path

mcu_files:
    dict[full_path] = (mtime, size)

host_files:
    host_path = os.path.join(package, file_path)
    dict[full_path] = (mtime, size, host_path)
"""

class EvalRsync(EvalRlist):
    """Add remote file synchronization"""

    def rsync(self, data_consumer, *,
            dry_run=True,
            upload_only=True):
        # synchronize micrcontroller flash to host
        #   dry_run: only print out differences, do not copy any files
        #   upload_only: do not delete files on microcontroller that are not also on host
        if not dry_run:
            # sync mcu time to host if they differ by more than 3 seconds
            self.sync_time(3)
        # mcu files & excludes
        mcu_files = self.rlist('/', data_consumer)
        mcu_files.pop("/boot_out.txt", None)
        # host files
        host_files = self._host_files()
        del_, add_, upd_ = self._diff(mcu_files, host_files)
        with cd(Config.iot49_dir()):
            same = True
            for dst_file in del_:
                # delete first (protect against a bug that deletes what was just copied)
                if not upload_only:
                    same = False
                    data_consumer(colored(f"DELETE  {dst_file}\n", 'red'))
                    if not dry_run:
                        self.rm_rf(dst_file)
            for dst_file, src_file in add_.items():
                # no feedback about directory creation
                if os.path.isfile(src_file):
                    same = False
                    data_consumer(colored(f"ADD     {dst_file}\n", 'green'))
                if not dry_run:
                    self.fput(src_file, dst_file)
            for dst_file, src_file in upd_.items():
                same = False
                data_consumer(colored(f"UPDATE  {dst_file}\n", 'blue'))
                if not dry_run:
                    self.fput(src_file, dst_file)
        if same:
            data_consumer(colored("Directories match\n", 'green'))

    def _diff(self, mcu_files, host_files):
        # determine difference between host (projects) and mcu
        # delete files not on host
        to_delete = mcu_files.keys() - host_files.keys()
        # add files from host
        to_add = host_files.keys() - mcu_files.keys()
        # in both: may need updating
        to_update = set()
        for u in mcu_files.keys() & host_files.keys():
            mcu_time, mcu_size = mcu_files[u]
            host_time, host_size, _ = host_files[u]
            # mcu_size < 0 indicates directory
            if mcu_size < 0: continue
            if mcu_size != host_size or mcu_time < host_time:
                to_update.add(u)
        # convert to_add and to_update to ordered dicts full_path --> host_path
        return (
            sorted(to_delete, reverse=True),
            OrderedDict(sorted({ k: host_files[k][-1] for k in to_add }.items())),
            OrderedDict(sorted({ k: host_files[k][-1] for k in to_update }.items()))
        )

    def _host_files(self):
        # returns { dict mcu_filename -> (mtime, size, host_filename) }
        result = {}
        # add folder so it won't be deleted
        result['/'] = (0, -1, '')
        pkgs = self.device.packages
        implementation = self.implementation
        with cd(Config.iot49_dir()):
            for pkg in pkgs:
                dest = Config.get_device(self.device.name).get_package_dest(pkg.name)
                # add folder so it won't be deleted
                # result[dest] = ('', 0, -1, '')
                for file_path, package in pkg.files().items():
                    src = os.path.join(package, file_path)
                    mtime = os.path.getmtime(src)
                    size = -1 if os.path.isdir(src) else os.path.getsize(src)
                    if file_path.endswith('.py'):
                        # check if we have a compiled version
                        mpy_src = os.path.join('.compiled', implementation, src.replace('.py', '.mpy'))
                        if os.path.isfile(mpy_src):
                            mpy_mtime = os.path.getmtime(mpy_src)
                            if mpy_mtime >= mtime:
                                # compiled file is newer
                                src = mpy_src
                                mtime = mpy_mtime
                                size = os.path.getsize(mpy_file)
                    result[os.path.normpath(os.path.join(dest, file_path))] = (int(mtime), size, src)
        return result



from .eval import RemoteError
from .device import Device
from .discover_serial import DiscoverSerial
from .discover_mdns import DiscoverMdns
from serial import SerialException
import os, logging, threading, time

logger = logging.getLogger(os.path.splitext(os.path.basename(__file__))[0])


class DeviceRegistry:
    """Keeps track of currently available devices"""

    def __init__(self):
        self._discover_serial = DiscoverSerial()
        self._discover_mdns = DiscoverMdns()
        # map url -> device
        self._devices = {}
        # map url -> time of last attempt
        self._registration_failed = {}

    @property
    def devices(self) -> frozenset:
        """List of all devices that are currently online. Includes InaccessibleDevice."""
        self._update()
        return frozenset(self._devices.values())

    def get_device(self, name: str, schemes=None) -> Device:
        """Device with given name/uid/url & schemes."""
        if schemes == None or len(schemes) == 0:
            schemes = ['serial', 'ws', 'mp']
        self._update()
        for scheme in schemes:
            for dev in self._devices.values():
                if (dev.name == name or dev.uid == name or dev.url == name) and dev.scheme == scheme:
                    return dev
        return None

    def _update(self):
        # update database ...
        urls = self._discover_serial.scan() | self._discover_mdns.scan()
        # 1) purge database of devices that are no longer available
        now = time.monotonic()
        for k in list(self._devices.keys()):
            if k in urls: continue
            v = self._devices[k]
            if not v.max_age: continue
            if (now-v.last_seen) > v.max_age:
                del self._devices[k]
        # 2) register newly discovered devices (not already in database)
        for url in urls:
            try:
                self.register(url, 0.5)
            except (ValueError, RemoteError) as e:
                # leave this print statement - will show up in jupyter!
                print(f"\nFailed to register {url}:\n{e}\n")
                logger.info(f"Failed to register {url}: {e}")

    def register(self, url:str, max_age:float=None):
        """Create device for given url and register in database.
        :param: max_age:float  Device automatically unregistered
                     if no activity (register) in specified interval.
                     Default: None, no automatic unregistering.
                     Calling register (repeatedly) resets age to 0.
        """
        if url in self._registration_failed.keys():
            if (time.monotonic() - self._registration_failed[url]) < 10:
                logger.info(f"skipping failed registration for '{url}'")
                return
            del self._registration_failed[url]
        if url in self._devices.keys():
            # already in database
            self._devices[url].last_seen = time.monotonic()
            return
        # create a new device
        logger.debug(f"register {url}")
        device_class = find_device_class(url)
        try:
            device = device_class(url)
        except Exception as e:
            self._registration_failed[url] = time.monotonic()
            msg = str(e)
            if url.startswith('ws'): msg += " (wrong password?)"
            print(f"Registration failed for {url}: {msg}")
            logger.error(f"Registration failed for {url}: {msg}")
            return
        device.max_age = max_age
        device.last_seen = time.monotonic()
        self._devices[url] = device
        logger.info(f"registered {device.name} {device.uid} {url}")

    def unregister(self, name:str):
        """Unregister device (by name, uid, or url)"""
        url = None
        for v in self._devices.values():
            if name == v.uid or name == v.name or name == v.url:
                url = v.url
                break
        if not url:
            raise ValueError(f"Device '{name}' not in registry")
        del self._devices[url]
        logger.debug(f"unregisted '{url}'")


def find_device_class(url):
    from .serial_device import SerialDevice
    from .mp_device import MpDevice
    from .telnet_device import TelnetDevice
    from .webrepl_device import WebreplDevice
    classes = {}
    classes['serial'] = SerialDevice
    classes['mp']     = MpDevice
    classes['ws']     = WebreplDevice
    classes['wss']    = WebreplDevice
    classes['telnet'] = TelnetDevice
    try:
        scheme, _ = url.split('://')
    except ValueError as e:
        raise ValueError(f"Invalid url: {url}")
    c = classes.get(scheme)
    if c: return c
    raise ValueError(f"Unknown scheme: {scheme}")



from .eval import Eval, RemoteError
from .config import Config

from abc import ABC, abstractmethod
import os, threading, time, logging

logger = logging.getLogger(os.path.splitext(os.path.basename(__file__))[0])


class Device(ABC):

    def __init__(self, url:str):
        """Create device for given URL.
        Retrieves uid from device and raises error if device cannot be contacted.

        :param: url:str
        e.g.
            serial:///dev/cu.usbmodem1413401
            telnet://mcu.com:23
        """
        self._url = url
        # connect to device and retrieve its uid; raise RemoteError if unsuccessful
        self._uid = url
        with self as repl:
            self._uid = repl.exec(_uid).decode()

    @abstractmethod
    def read(self, size=1) -> bytes:
        """Read bytes. Can be aborted by KeyboardInterrupt."""

    @abstractmethod
    def write(self, data: bytes):
        """Writes data to device."""

    @abstractmethod
    def inWaiting(self):
        """Number of bytes available for reading without blocking."""

    @abstractmethod
    def __enter__(self) -> Eval:
        """Usage pattern:

        with device as repl:
            repl.exec(...)
            repl.rlist(...)
        """

    @abstractmethod
    def __exit__(self, type, value, traceback):
        pass

    def __repr__(self):
        return f"{self.name} @ {self.url} ({self.uid})"

    @property
    def uid(self) -> str:
        """Cache"""
        return self._uid

    @property
    def url(self) -> str:
        return self._url

    @property
    def scheme(self) -> str:
        return self.url.split('://')[0]

    @property
    def address(self) -> str:
        return self.url.split('://')[1]

    @property
    def name(self) -> str:
        """Device name"""
        try:
            return Config.get_device(self.uid).name
        except ValueError:
            return self.uid

    @property
    def packages(self):
        try:
            return Config.get_device(self.uid).get_packages()
        except ValueError:
            return []


###############################################################################
# code snippet (run on remote)

_uid = """
uid = bytes(6)
try:
    import machine
    uid = machine.unique_id()
except:
    import microcontroller
    uid = microcontroller.cpu.uid
print(":".join("{:02x}".format(x) for x in uid), end="")
"""



from glob import glob
from fnmatch import fnmatch
from types import ModuleType
import inspect, io, os, logging

logger = logging.getLogger('config')


class Config:

    @staticmethod
    def iot49_dir():
        return os.path.expanduser(os.getenv('IOT49', '~/iot49'))

    @staticmethod
    def get_attributes():
        def filt(elem):
            k, v = elem
            skip = isinstance(v, ModuleType) or callable(v) or k == 'logger'
            return not skip
        cfg = Config.get_config()
        del cfg['__builtins__']
        return dict(filter(filt, cfg.items()))

    @staticmethod
    def get_attr(name, default=None):
        c = Config.get_config()
        return c.get(name, default)

    @staticmethod
    def get_packages():
        c = Config.get_config()
        return c['Package'].get_packages()

    @staticmethod
    def get_package(name):
        c = Config.get_config()
        return c['Package'].get_package(name)

    @staticmethod
    def get_devices():
        c = Config.get_config()
        return c['Device'].get_devices()

    @staticmethod
    def get_device(name_or_uid):
        c = Config.get_config()
        return c['Device'].get_device(name_or_uid)

    @staticmethod
    def get_config():
        """Load configuration from cache or disk."""
        from .utilities import cd
        from . import config
        cfg = {}
        src = inspect.getsource(config)
        exec(src, cfg)
        file = 'config'
        try:
            with cd(os.path.join(Config.iot49_dir(), 'config')):
                for file in glob('./**/*.py', recursive=True):
                    with open(file) as f:
                        try:
                            exec(f.read(), cfg)
                        except SyntaxError as e:
                            raise SyntaxError(f"{e.text.strip()}: {e.msg}, line {e.lineno} file {file}")
        except (NameError, OSError) as ne:
            logger.error("{} while reading {}".format(ne, file))
            raise
        except SyntaxError as se:
            logger.error("{} in {}".format(se, file))
            raise
        return cfg


class Package:

    # dict name -> Package
    __PACKAGES = {}

    @staticmethod
    def get_packages():
        return Package.__PACKAGES

    @staticmethod
    def get_package(name):
        if '/' in name:
            # path, not package name: create package on the fly
            return Package(name, name)
        return Package.__PACKAGES.get(name)

    def __init__(self, name, src, requires=[], includes=None, excludes=None):
        # src: list of dirs or files relative to $IOT49
        # may also the name of a single dir or file
        if name in Package.__PACKAGES:
            raise ValueError(f"Redefinition of Package '{name}'")
        self._name = name
        if isinstance(src, str): src = [src]
        self._src = src
        if isinstance(requires, str): requires = [requires]
        self._requires = requires
        self._includes = includes or ['./**/*.py', './**/*.mpy', './**/']
        self._excludes = excludes or ['boot_out.txt']
        # don't register "on the fly" packages
        if not '/' in name:
            Package.__PACKAGES[name] = self

    @property
    def name(self):
        return self._name

    def files(self):
        result = {}  # file -> path
        for req in self._requires:
            result.update(self.get_package(req).files())
        for src in self._src:
            path = os.path.join(Config.iot49_dir(), src)
            if os.path.isfile(path):
                p_, f_ = os.path.split(src)
                result[f_] = p_
                continue
            if not os.path.isdir(path):
                logger.warn(f"Folder/file '{path}' not found (ignored)")
                continue
            from iot_device import cd
            with cd(path):
                for inc in self._includes:
                    for file in glob(inc, recursive=True):
                        for ex in self._excludes:
                            if fnmatch(file, ex): continue
                        file = os.path.normpath(file)
                        result[file] = src
        return result

    def description(self):
        s = io.StringIO()
        s.write(f"Package {self._name}:\n")
        files = dict(sorted(self.files().items(), key=lambda item: item[0]))
        for a, b in files.items():
            s.write(f"  {a:40} from {b}\n")
        return s.getvalue()

    def __str__(self):
        return self._name



class Device:

    # dict name --> device
    __DEVICES = {}

    @staticmethod
    def get_devices():
        return Device.__DEVICES

    @classmethod
    def get_device(cls, name):
        dev = cls.__DEVICES.get(name)
        if dev: return dev
        # search for uid
        for dev in cls.__DEVICES.values():
            if dev.uid == name: return dev
        raise ValueError(f"No configuration for device '{name}'")

    def __init__(self, name, uid, **kwargs):
        self._dict = {
            'name': name,
            'uid': uid,
        }
        self._dict.update(kwargs)
        Device.__DEVICES[name] = self

    def get_packages(self):
        names = self._dict.get('packages', [])
        if isinstance(names, str): names = [names]
        res = []
        for name in names:
            if not isinstance(name, str): name = name[0]
            p = Package.get_package(name)
            if p:
                res.append(p)
            else:
                logger.warn(f"No such package: '{name}' (skipped)")
        return res

    def get_package_dest(self, pkg_name):
        def root(name): return name if name.startswith('/') else '/' + name
        names = self._dict.get('packages', [])
        for name in names:
            if isinstance(name, str): continue
            if name[0] == pkg_name: return root(name[1])
        return root(self._dict.get('dest', ''))

    def __getattr__(self, name):
        return self._dict.get(name)

    def description(self):
        s = io.StringIO()
        s.write(f"Device {self._dict.get('name')}:")
        for k,v in self._dict.items():
            if k == 'name': continue
            s.write(f"    {k} = {v}\n")
        return s.getvalue()

    def __str__(self):
        return f"{self._dict.get('name')} ({self._dict.get('uid')})"



from .discover import Discover
from .eval import RemoteError

from serial import SerialException
import serial.tools.list_ports
import os, time, logging, threading

logger = logging.getLogger(os.path.splitext(os.path.basename(__file__))[0])


# Vendor IDs
ADAFRUIT_VID = 0x239A  # Adafruit board
PARTICLE_VID = 0x2B04  # Particle
ESP32_VID    = 0x10C4  # ESP32 via CP2104
STM32_VID    = 0xf055  # STM32 usb port

COMPATIBLE_VID = { ADAFRUIT_VID, PARTICLE_VID, STM32_VID, ESP32_VID }


class DiscoverSerial(Discover):

    def __init__(self, scan_rate:float=1):
        """Start a daemon thread that continually scans ports every scan_rate seconds."""
        super().__init__()

    def scan(self):
        # return url's of devices that are online
        res = set()
        for port in serial.tools.list_ports.comports():
            if port.vid in COMPATIBLE_VID:
                res.add(f"serial://{port.device}")
            elif port.vid:
                logger.info(f"Found {port} with unknown VID {port.vid:02X} (ignored)")
        return res



from .device_registry import DeviceRegistry
from .certificate import create_key_cert_pair
from .config import Config

from zeroconf import ServiceInfo, Zeroconf
from serial import SerialException
import os
import socket
import selectors
import ssl
import threading
import json
import time
import tempfile
import logging

logger = logging.getLogger(os.path.splitext(os.path.basename(__file__))[0])


def _my_ip():
    # determine host's ip address
    with socket.socket(socket.AF_INET, socket.SOCK_DGRAM) as s:
        # fake address, does not need to be reachable
        s.connect(('10.1.1.1', 1))
        return s.getsockname()[0]


class AdvertiseServer:

    def __init__(self):
        self._zeroconf = Zeroconf()
        self._addresses = [socket.inet_aton(_my_ip())]
        self._id2info = {}

    def register_device(self, id, device):
        info = ServiceInfo(
            type_="_repl._tcp.local.",
            name=device.name + "." + "_repl._tcp.local.",
            port=Config.get_attr('server_port', '34567'),
            properties = { "uid": device.uid, "name": device.name },
            addresses=self._addresses)
        self._id2info[id] = info
        self._zeroconf.register_service(info, allow_name_change=True)
        logger.debug(f"advertise {id}")

    def unregister_device(self, id):
        try:
            logger.debug(f"no longer advertise {id}")
            info = self._id2info[id]
            if info:
                self._zeroconf.unregister_service(info)
                del self._id2info[id]
        except AttributeError:
            pass


class DeviceServer():

    def __init__(self):
        # serve devices in device_registry
        self.__ip = _my_ip()
        self.__ssl_context = self.__make_ssl_context()

    def serve(self):
        """Serve multiple connections to different devices in parallel. Never returns."""
        self.__sel = selectors.DefaultSelector()
        lsock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        lsock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        port = Config.get_attr('server_port', '34567')
        lsock.bind(('', port))
        lsock.listen()
        logger.info(f"Listening for connections on {self.__ip}:{port}")
        lsock.setblocking(False)
        self.__sel.register(lsock, selectors.EVENT_READ, data=None)
        while True:
            try:
                events = self.__sel.select(timeout=None)
                for key, mask in events:
                    if key.data is None:
                        # connect request (key.fileobj is socket)
                        self.__accept_wrapper(key.fileobj)
                    else:
                        # client connection
                        self.__service_connection(key, mask)
            except Exception as e:
                logger.exception("unhandled exception")


    def __accept_wrapper(self, sock):
        # accept connection
        conn, addr = sock.accept()
        logger.debug(f"accept connection at {(conn, addr)}")
        conn = self.__ssl_context.wrap_socket(conn, server_side=True)
        # More quickly detect bad clients who quit without closing the
        # connection: After 1 second of idle, start sending TCP keep-alive
        # packets every 1 second. If 3 consecutive keep-alive packets
        # fail, assume the client is gone and close the connection.
        try:
            conn.setsockopt(socket.IPPROTO_TCP, socket.TCP_KEEPIDLE, 1)
            conn.setsockopt(socket.IPPROTO_TCP, socket.TCP_KEEPINTVL, 1)
            conn.setsockopt(socket.IPPROTO_TCP, socket.TCP_KEEPCNT, 3)
            conn.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1)
        except AttributeError:
            pass  # not available on windows
        conn.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
        # get uid and password (blocking recv)
        # TODO: check for incomplete message!
        try:
            uid_pwd = json.loads(conn.recv(1024).decode())
        except ssl.SSLWantReadError:
            logger.error("SSLWantReadError in device_server.__accept_wrapper")
        except (ConnectionResetError, Exception):
            conn.close()
            return
        uid = uid_pwd.get('uid', '?')
        device = DeviceRegistry.get_device(uid)
        logger.debug(f"Request from {addr} to {uid}")
        # check password & device status
        ans = None
        if uid_pwd.get('password') != Config.get_attr('password', '?'):
            ans = f'wrong password for {uid}'
        elif not device:
            ans = f'no device {uid}'
        if ans:
            conn.write(ans.encode())
            conn.close()
        else:
            conn.write(b'ok')
            conn.setblocking(False)
            device.__enter__()
            events = selectors.EVENT_READ | selectors.EVENT_WRITE
            self.__sel.register(conn, events, data=device)

    def __service_connection(self, key, mask):
        try:
            sock = key.fileobj
            device = key.data
            if mask & selectors.EVENT_READ:
                # debug code for dealing with SSLWantReadError
                # from Python docs:
                #    exception ssl.SSLWantReadError
                #    A subclass of SSLError raised by a non-blocking SSL socket when
                #    trying to read or write data, but more data needs to be received
                #    on the underlying TCP transport before the request can be fulfilled.
                for _ in range(10):
                    try:
                        recv_data = sock.recv(256)
                        break
                    except ssl.SSLWantReadError:
                        logger.warning("SSLWantReadError in device_server.__service_connection")
                        time.sleep(0.5)
                if recv_data:
                    device.write(recv_data)
                else:
                    logger.info(f"Closing connection to {device.uid}")
                    self.__sel.unregister(sock)
                    sock.close()
                    device.__exit__(None, None, None)
            if mask & selectors.EVENT_WRITE:
                # forward data from device, if any
                msg = device.read_all()
                if len(msg) > 0:
                    sock.sendall(msg)
        except (SerialException, ConnectionResetError, OSError) as e:
            logger.info(f"Communication with {device.uid} failed, closing connection ({e})")
            self.__sel.unregister(sock)
            sock.close()
            device.__exit__(None, None, None)

    @staticmethod
    def __make_ssl_context():
        context = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
        with tempfile.NamedTemporaryFile() as cert_file:
            key, cert = create_key_cert_pair()
            cert_file.write(key)
            cert_file.write(cert)
            cert_file.seek(0)
            context.load_cert_chain(certfile=cert_file.name)
        context.options |= ssl.OP_NO_TLSv1 | ssl.OP_NO_TLSv1_1
        context.set_ciphers('EECDH+AESGCM:EDH+AESGCM:AES256+EECDH:AES256+EDH')
        return context


##########################################################################
# Main

def main():
    from .discover_serial import DiscoverSerial
    import sys

    logging.basicConfig()
    logging.getLogger('device_server').setLevel(logging.INFO)

    # scan serial ports and advertise
    DiscoverSerial()
    DeviceRegistry.register_listener(AdvertiseServer())

    # accept connections and device communication
    DeviceServer().serve()


if __name__ == "__main__":
    main()



from .device import Device
from .eval import RemoteError
from .repl_protocol import ReplProtocol

from serial import Serial, SerialException
import os, time, logging

logger = logging.getLogger(os.path.splitext(os.path.basename(__file__))[0])

class SerialDevice(Device):

    def __init__(self, url):
        super().__init__(url)

    def read(self, size):
        return self.__serial.read(size)

    def write(self, data):
        return self.__serial.write(data)

    def inWaiting(self):
        return self.__serial.in_waiting

    def __enter__(self):
        try:
            self.__serial = Serial(self.address, 115200, parity='N',
                timeout=0.5,            # read timeout
                write_timeout=2,
                exclusive= True         # exclusive access mode (POSIX only)
            )
            return ReplProtocol(self)
        except (BlockingIOError, SerialException):
            raise
            # raise RemoteError(f"Device {self.url} not available (in use?)")
        except Exception as e:
            raise RemoteError(f"Device {self.url} encountered problem: {e}")

    def __exit__(self, type, value, traceback):
        self.__serial.close()
        self.__serial = None



from .eval import RemoteError
from .eval_rsync import EvalRsync
from .pyboard import PyboardError
from .pydevice import Pydevice
from websocket import WebSocketException
import logging, os, time

logger = logging.getLogger(os.path.splitext(os.path.basename(__file__))[0])

class ReplProtocol(EvalRsync):
    """Concrete class of Eval (implements exec etc)"""

    def __init__(self, device):
        super().__init__(device)
        self.pyboard = Pydevice(device)
        self.pyboard.enter_raw_repl()

    def close(self):
        self.pyboard.exit_raw_repl()

    def exec(self, code: str, data_consumer=None) -> bytes:
        try:
            res = self.pyboard.exec(code, data_consumer)
            # Don't try to use raw-paste mode again unless supported by this device
            self.device.use_raw_paste = self.pyboard.use_raw_paste
            return res
        except OSError as e:
            raise RemoteError("Device disconnected")
        except PyboardError as e:
            raise RemoteError(*e.args)

    def softreset(self) -> None:
        try:
            self.pyboard.softreset()
        except WebSocketException as e:
            # Connection is already closed
            # webrepl needs some time to restart
            time.sleep(3)
        except PyboardError as e:
            logger.exception(f"softreset: {e}")
            raise RemoteError(*e.args)

    def abort(self) -> None:
        try:
            # enter_raw_repl aborts running program (if any)
            self.pyboard.enter_raw_repl()
        except PyboardError as e:
            logger.exception("abort: {e}")
            raise RemoteError(*e.args)



from .device import Device
from .eval import RemoteError
from .config import Config
from .mp_protocol import MpProtocol

import os, socket, time, logging

logger = logging.getLogger(os.path.splitext(os.path.basename(__file__))[0])

VERSION = 'ns-V1'

class PasswordError(Exception):
    pass

class MpDevice(Device):

    def __init__(self, url):
        super().__init__(url)

    def __enter__(self):
        assert self.scheme == "mp"
        addr, port = self.address.split(':')
        self.sock = s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        logger.debug(f"connect ({addr}, {port})")
        s.settimeout(1)
        s.connect((addr, int(port)))
        version = self.readline()
        logger.debug(f"version {version}")
        if version == '': raise RemoteError(f"Device is offline")
        if version != VERSION: raise RemoteError(f"Wrong mp version: client={repr(VERSION)}, server={repr(version)}")
        s.sendall(Config.get_attr("mp_pwd", "?").encode())
        s.sendall(b'\n')
        ok = self.readline()
        logger.debug(f"ok = {ok}")
        if ok != 'OK': raise RemoteError(f"Expected OK, got {ok}")
        return MpProtocol(self)

    def __exit__(self, typ, value, traceback):
        self.sock.sendall(b'bye\n')
        try:
            self.sock.close()
        except:
            pass
        self.sock = None

    def readline(self, timeout=1):
        """Blocking"""
        res = b''
        start = time.monotonic()
        # let's give the MCU some time ...
        while (time.monotonic() - start) < timeout:
            try:
                b = self.sock.recv(1)
            except socket.timeout:
                time.sleep(0.1)
                continue
            if b == b'':
                raise Exception("connection closed")
            if b == b'\n':
                return res.decode()
            res += b
        return res.decode()

    def read(self, size):
        raise RuntimeError("MpDevice read()")

    def write(self, data):
        raise RuntimeError("MpDevice write()")

    def inWaiting(self):
        raise RuntimeError("MpDevice inWaiting()")



from abc import ABC, abstractmethod
import os, inspect, logging, time

logger = logging.getLogger(os.path.splitext(os.path.basename(__file__))[0])


class RemoteError(Exception):
    def __init__(self, msg, output=None, error=None):
        # print(f"RemoteError:\n  msg = {msg}\n  out = {output}\n  err = {error}")
        super().__init__(msg, output, error)

    @property
    def exception(self):
        try:
            return self.args[2].strip().split('\n')[-1]
        except:
            return self.args[0]

    @property
    def msg(self):
        return self.args[0] or b''

    @property
    def output(self):
        return self.args[1] or b''

    @property
    def traceback(self):
        return self.args[2] or b''

    def __str__(self):
        if self.traceback:
            try:
                return self.traceback.decode()
            except UnicodeError:
                return self.traceback
        return self.msg


def default_data_consumer(data:bytes):
    print(data)


class Eval(ABC):
    """Abstract class encapsulating code evaluation on microcontroller."""

    def __init__(self, device):
        self._device = device

    @property
    def device(self):
        return self._device

    @abstractmethod
    def exec(self, code: str, data_consumer=None) -> bytes:
        """Exec code on remote (Micro)Python VM.
        If data_consumer is not None, output (from print statements) is
        passed to it as it becomes available.
        Otherwise, output is collected and returned as as bytes.
        """

    @abstractmethod
    def abort(self) -> None:
        """Abort currently running program without resetting the MicroPython VM"""

    @abstractmethod
    def softreset(self) -> None:
        """Reset MicroPython VM."""



import os
import re
import setuptools
from iot_device import version

install_requires = [
    "pyserial",
    "termcolor",
    "pyopenssl",
    "zeroconf",
    "websocket-client",
]

setuptools.setup(
    name="iot-device",
    version=version.__version__,
    packages=[ 'iot_device' ],
    author="Bernhard Boser",
    description="MicroPython REPL Interface",
    long_description="Communication with MicroPython device over serial, internet, ...",
    long_description_content_type="text/markdown",
    license="MIT",
    keywords="MicroPython,Repl",
    url="https://github.com/iot49/iot-device",
    classifiers = [
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
    ],
    install_requires=install_requires,
    include_package_data=True,
    python_requires='>=3.7',
    zip_safe = True,
)



from .device import Device
from .eval import RemoteError
from .repl_protocol import ReplProtocol

from telnetlib import Telnet
from collections import deque
import os, select, time, logging

logger = logging.getLogger(os.path.splitext(os.path.basename(__file__))[0])

class TelnetDevice(Device):

    def __init__(self, url):
        self.use_raw_paste = False
        self.fifo = deque()
        self.read_timeout = 10
        super().__init__(url)

    def read(self, size=1):
        while len(self.fifo) < size:
            timeout_count = 0
            data = self.__telnet.read_eager()
            if len(data):
                self.fifo.extend(data)
                timeout_count = 0
            else:
                time.sleep(0.1)
                if self.read_timeout is not None and timeout_count > 4 * self.read_timeout:
                    break
                timeout_count += 1

        data = b""
        while len(data) < size and len(self.fifo) > 0:
            data += bytes([self.fifo.popleft()])
        return data

    def write(self, data):
        if len(data) < 220:
            return self.__telnet.write(data)
        # slow to avoid communication errors
        # (esp32 notify buffer overflow)
        chunk_size = 64
        n = 0
        for i in range(0, len(data), chunk_size):
            n += self.__telnet.write(data[i:min(i+chunk_size, len(data))])
            if n < len(data):
                time.sleep(0.2)
        return n

        self.__telnet.write(data)
        return len(data)

    def inWaiting(self):
        n_waiting = len(self.fifo)
        if not n_waiting:
            data = self.__telnet.read_eager()
            self.fifo.extend(data)
            return len(data)
        else:
            return n_waiting

    def __enter__(self):
        addr, port = self.address.split(':')
        self.__telnet = Telnet(addr, port, timeout=15)
        print(f"TelnetDevice({addr}, {port}) - enter")
        return ReplProtocol(self)

    def __exit__(self, type, value, traceback):
        try:
            print("TelnetDevice.close")
            self.__telnet.close()
            print("closed")
        except Exception as e:
            print("*** close", e)
            pass
        self.__telnet = None



from abc import ABC, abstractmethod


class Discover(ABC):
    """Base class for device discovery"""

    def __init__(self):
        pass

    @abstractmethod
    def scan(self) -> list:
        """url's of devices that are online"""
